/* -*- mode: java; c-basic-offset: 2; indent-tabs-mode: nil -*- */

/*
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package processing.app;

import processing.app.debug.MessageConsumer;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.swing.text.*;
import java.io.*;
import java.net.*;

public class SerialMonitor extends JFrame implements MessageConsumer {
  private Serial serial;
  private String port;
  private JTextArea textArea;
  private JScrollPane scrollPane;
  private JTextField textField;
  private JButton sendButton;
  private JCheckBox autoscrollBox;
  private JComboBox lineEndings;
  private JComboBox serialRates;
  private int serialRate;

  public SerialMonitor(String port) {
    super(port);
  
    this.port = port;
  
    addWindowListener(new WindowAdapter() {
        public void windowClosing(WindowEvent e) {
          closeSerialPort();
        }
      });  
      
    // obvious, no?
    KeyStroke wc = Editor.WINDOW_CLOSE_KEYSTROKE;
    getRootPane().getInputMap(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT).put(wc, "close");
    getRootPane().getActionMap().put("close", new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        closeSerialPort();
        setVisible(false);
      }});
  
    getContentPane().setLayout(new BorderLayout());
    
    Font font = Theme.getFont("console.font");

    textArea = new JTextArea(16, 40);
    textArea.setEditable(false);    
    textArea.setFont(font);
    
    // don't automatically update the caret.  that way we can manually decide
    // whether or not to do so based on the autoscroll checkbox.
    ((DefaultCaret)textArea.getCaret()).setUpdatePolicy(DefaultCaret.NEVER_UPDATE);
    
    scrollPane = new JScrollPane(textArea);
    
    getContentPane().add(scrollPane, BorderLayout.CENTER);
    
    JPanel pane = new JPanel();
    pane.setLayout(new BoxLayout(pane, BoxLayout.X_AXIS));
    pane.setBorder(new EmptyBorder(4, 4, 4, 4));

    textField = new JTextField(40);
    textField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        send(textField.getText());
        textField.setText("");
      }});

    sendButton = new JButton("Send");
    sendButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        send(textField.getText());
        textField.setText("");
      }});
    
    pane.add(textField);
    pane.add(Box.createRigidArea(new Dimension(4, 0)));
    pane.add(sendButton);
    
    getContentPane().add(pane, BorderLayout.NORTH);
    
    pane = new JPanel();
    pane.setLayout(new BoxLayout(pane, BoxLayout.X_AXIS));
    pane.setBorder(new EmptyBorder(4, 4, 4, 4));
    
    autoscrollBox = new JCheckBox("Autoscroll", true);
    
    lineEndings = new JComboBox(new String[] { "No line ending", "Newline", "Carriage return", "Both NL & CR" });
    lineEndings.setMaximumSize(lineEndings.getMinimumSize());
      
    String[] serialRateStrings = {
      "300","1200","2400","4800","9600","14400",
      "19200","28800","38400","57600","115200"
    };
    
    serialRates = new JComboBox();
    for (int i = 0; i < serialRateStrings.length; i++)
      serialRates.addItem(serialRateStrings[i] + " baud");

    serialRate = Preferences.getInteger("serial.debug_rate");
    serialRates.setSelectedItem(serialRate + " baud");
    serialRates.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        String wholeString = (String) serialRates.getSelectedItem();
        String rateString = wholeString.substring(0, wholeString.indexOf(' '));
        serialRate = Integer.parseInt(rateString);
        Preferences.set("serial.debug_rate", rateString);
        closeSerialPort();
        try {
	        openSerialPort();
        } catch (SerialException e) {
          System.err.println(e);
        }
      }});
      
    serialRates.setMaximumSize(serialRates.getMinimumSize());

    pane.add(autoscrollBox);
    pane.add(Box.createHorizontalGlue());
    pane.add(lineEndings);
    pane.add(Box.createRigidArea(new Dimension(8, 0)));
    pane.add(serialRates);
    
    getContentPane().add(pane, BorderLayout.SOUTH);

    pack();
  }
  
  private void send(String s) {
    if (serial != null) {
      switch (lineEndings.getSelectedIndex()) {
        case 1: s += "\n"; break;
        case 2: s += "\r"; break;
        case 3: s += "\r\n"; break;
      }
      serial.write(s);
    }
  }
  
  public void openSerialPort() throws SerialException {
    if (serial != null) return;
  
    if (Base.getBoardPreferences().get("serial.restart_cmd") != null) {
      RestartCommand r = new RestartCommand(port);
      r = null;
    }
    String fake = Base.getBoardMenuPreference("fake_serial");
    if (fake == null) {
      if (Base.getBoardPreferences().get("serial.safe_baud_rates_only") != null) {
        if (serialRate == 14400) serialRate = 19200;
        if (serialRate == 28800) serialRate = 38400;
      }
      serial = new Serial(port, serialRate);
    } else {
      serial = new FakeSerial(fake);
    }
    serial.addListener(this);
  }
  
  public void closeSerialPort() {
    if (serial != null) {
      textArea.setText("");
      serial.dispose();
      serial = null;
    }
  }
  
  public void message(final String s) {
    SwingUtilities.invokeLater(new Runnable() {
      public void run() {
        textArea.append(s);
        if (autoscrollBox.isSelected()) {
        	textArea.setCaretPosition(textArea.getDocument().getLength());
        }
      }});
  }
}


class FakeSerial extends Serial {
       Socket sock=null;
       inputListener listener=null;
       int[] addrlist = {28541,4984,18924,16924,27183,31091};
       Process gateway=null;

       public FakeSerial(String name) throws SerialException {
               super("fake serial");
               int attempt=1;
               do {
                       if (gateway_connect(name)) return;
                       if (attempt == 1 && !gateway_start(name)) {
                               System.err.println("Error starting " + name);
                       }
                       delay_20ms();
               } while (++attempt < 4);
               throw new SerialException("no connection");
       }

       private boolean gateway_connect(String name) {
               int namelen = name.length();
               byte[] buf = new byte[namelen];
               byte[] namebuf = name.getBytes();
               for (int i=0; i<addrlist.length; i++) {
                       try {
                               sock = new Socket("127.0.0.1", addrlist[i]);
                               input = sock.getInputStream();
                               output = sock.getOutputStream();
                       } catch (Exception e) {
                               continue;
                       }
                       // check for welcome message
                       try {
                               int wait = 0;
                               while (input.available() < namelen) {
                                       if (++wait > 6) throw new Exception();
                                       delay_20ms();
                               }
                               input.read(buf, 0, namelen);
                               String id = new String(buf, 0, namelen);
                               for (int n=0; n<namelen; n++) {
                                       if (buf[n] !=  namebuf[n]) throw new Exception();
                               }
                       } catch (Exception e) {
                               // mistakenly connected to some other program!
                               close_sock();
                               continue;
                       }
                       return true;
               }
               sock = null;
               return false;
       }

       private void close_sock() {
               try {
                       sock.close();
               } catch (Exception e) { }
               sock = null;
       }

       private void delay_20ms() {
               try {
                       Thread.sleep(20);
               } catch (Exception e) { }
       }

       private boolean gateway_start(String cmd) {
               String path;
               if (Base.isLinux()) {
                       path = Base.getHardwarePath() + "/tools/";
               } else {
                       path = Base.getHardwarePath() + "/tools/avr/bin/";
               }
               try {
                       gateway = Runtime.getRuntime().exec(path + cmd);
               } catch (Exception e) {
                       return false;
               }
               return true;
       }

       public void dispose() {
               if (listener != null) {
                       listener.interrupt();
                       listener.consumer = null;
                       listener = null;
               }
               if (sock != null) {
                       try {
                               sock.close();
                       } catch (Exception e) { }
                       sock = null;
               }
       }

       public void addListener(MessageConsumer c) {
               if (sock == null) return;
               if (listener != null) listener.interrupt();
               listener = new inputListener();
               listener.input = input;
               listener.consumer = c;
               listener.start();
       }

       public void write(byte bytes[]) {
               if (output == null) return;
               if (bytes.length > 0) {
                       try {
                               output.write(bytes, 0, bytes.length);
                       } catch (IOException e) { }
               }
       }
       public void setDTR(boolean state) {
       }
       static public String[] list() {
               return null;
       }
       protected void finalize() throws Throwable {
               try {
                       try {
                               if (listener != null && listener.isAlive()) listener.interrupt();
                       } catch (Exception e) {};
                       try {
                               if (sock != null && sock.isConnected()) sock.close();
                       } catch (Exception e) {};
                       if (gateway != null) gateway.destroy();
               } finally {
                       super.finalize();
               }
       }
}

class RestartCommand{
	private Process restarter=null;
	public RestartCommand(String port) {
		if (Base.getBoardMenuPreference("fake_serial") == null) {
			try {
				Serial s = new Serial(port, 150);
				s.dispose();
				s = null;
			} catch (SerialException e) {
			}
		} else {
			String path;
			if (Base.isLinux()) {
				path = Base.getHardwarePath() + "/tools/";
			} else {
				path = Base.getHardwarePath() + "/tools/avr/bin/";
			}
			try {
				restarter = Runtime.getRuntime().exec(path + "teensy_restart");
			} catch (Exception e) {
			}
		}
	}
}

class inputListener extends Thread {
       MessageConsumer consumer;
       InputStream input;

       public void run() {
               byte[] buffer = new byte[1024];
               int num, errcount=0;
               try {
                       while (true) {
                               try {
                                       num = input.read(buffer);
                               } catch (IOException e) {
                                       if (isInterrupted()) break;
                                       num = -1;
                               }
                               if (num > 0) {
                                       errcount = 0;
                                       consumer.message(new String(buffer, 0, num));
                               } else {
                                       if (++errcount > 10) {
                                               System.err.println("teensy gateway communication error");
                                               break;
                                       } else {
                                               try {
                                                       Thread.sleep(10);
                                               } catch (Exception e) {
                                                       if (isInterrupted()) break;
                                               }
                                       }
                               }
                       }
               } catch (Exception e) { }
       }
}

